from unitgrade import hide, UTestCase
from exam_generator.exam import Question, jinjafy
from types import SimpleNamespace
import gym
import numpy as np
import sympy as sym
import os
import inspect
from unitgrade import NotebookTestCase
from exam_generator.mat1.mat1exam import Mat1Question
from exam_generator.mat1.mat1exam import QuestionController

class Sympy1(Mat1Question):
    notebook = 'homework.ipynb'
    sheet = """
## A problem with a derivative

### a) This problem will test your awesome differentiation skills.

Consider the function $f(x) = {{latex}}$. What $y_1 = f'({{val}})$? 
```{code-cell} ipython3
y1 = None  # Write your result here.
y1 = {{derivative}} #!b;warn #!b
```
"""
    def test_derivative(self):
        self.assertIsNotNone(self.nb.y1)

    @hide
    def test_derivative_hidden(self):
        self.assertEqualC(self.nb.y1)

    def generate(self):
        x = SimpleNamespace()
        x.a = 1+np.random.randint(4)
        x.b = 1+np.random.randint(2)
        x.val = np.random.randint(5)+2
        import sympy as sym
        z = sym.symbols('x')
        from sympy import latex
        f = x.a * z + z ** x.b
        x.latex = latex(f)
        x.derivative = sym.diff(f, z).subs(z, x.val)
        return x

class Sympy2(Mat1Question):
    notebook = 'homework.ipynb'
    sheet = """
## A problem with more troublesome derivative

### a) Solving an equation involving a derivative

Consider the function $f(x) = {{latex}}$. Suppose that ${{derivative}}= f'({{val}})$, what is $a$?
```{code-cell} ipython3
a = None  # Write your result here.
a = {{a}} #!b;warn #!b
```
"""
    def test_derivative(self):
        self.assertIsNotNone(self.nb.a)

    @hide
    def test_derivative_hidden(self):
        self.assertEqualC(self.nb.a)

    def generate(self):
        d = SimpleNamespace()
        d.a = 3+np.random.randint(6)
        d.b = 2+np.random.randint(10)
        d.val = np.random.randint(5)+2
        import sympy as sym
        x, a = sym.symbols(['x', 'a'])
        from sympy import latex
        f = d.b/a * x**a
        d.latex = latex(f)
        d.derivative = sym.diff(f, x).subs(x, d.val).subs(a, d.a)
        return d




class SymSinCos(Mat1Question):
    notebook = 'homework.ipynb'
    sheet = """
## Derivatives of trigonometric functions

### a) A derivative of the sinus function

Consider the function $f(x) = {{latex}}$. What is $f'({{x0}})$?
```{code-cell} ipython3
q2 = None  # Write your result here.
q2 = {{sol}} #!b;warn #!b
```
"""
    def test_derivative(self):
        self.assertIsNotNone(self.nb.q2)

    @hide
    def test_derivative_hidden(self):
        self.assertEqualC(self.nb.q2)

    def generate(self):
        d = SimpleNamespace()
        d.a = 3+np.random.randint(6)
        d.b = 2+np.random.randint(10)
        d.x0 = np.random.randint(5)+2
        import sympy as sym
        x, a = sym.symbols(['x', 'a'])

        from sympy import latex
        f = sym.sin(d.a * x) * d.b
        d.latex = latex(f)

        # sym.diff(f, x).subs(x, d.x0)

        d.sol = float(sym.diff(f, x).subs(x, d.x0) )

        return d



class W1Q2Controller(QuestionController):
    questions = [Sympy1, Sympy2, SymSinCos]


if __name__ == "__main__":
    q = SymSinCos(seed=3)
    q.render_dummy(r = 3)



import os.path

from exam_generator.exam import Question, MCQuestion
from types import SimpleNamespace
import numpy as np

def uroll(bank):
    options = {'true': [], 'false': []}
    # ft = []
    # tt = []
    for k in bank:
        if k in ['true', 'false']:
            options[k] += bank[k]
        else:
            t, f = uroll(bank[k])
            options['true'] += t
            options['false'] += f
    return options['true'], options['false']


class ExplicitMultipleChoice(MCQuestion):
    def __init__(self, seed, title='', template_base='explicit_question', clearpage=False, **kwargs):
        if template_base.endswith(".tex"):
            raise Exception("Template base should not end with .tex:", template_base)

        super().__init__(seed, title=title, clearpage=clearpage, **kwargs)
        self.template_base = template_base
        # self.bank = bank

    def render(self, exam):
        dirs = exam.get_dirs()

        def chk(path):
            dn = os.path.dirname(dirs.base +"/"+path)
            if not os.path.isdir(dn):
                os.mkdir(dn)
            if not os.path.isfile(dirs.base + "/" + path):
                with open(dirs.base + "/" + path, 'w') as f:
                    if path.endswith("_body.tex"):
                        s = "Which one of the following options are correct?\n"
                        s += """ 
\\iffalse
\\begin{comment} % A should be correct.
\\begin{itemize} 
    \\item Answer A
    \\item Answer B
    \\item Answer C
    \\item Answer D
\\end{itemize}
\\end{comment}
\\fi
\\iffalse
\\begin{solution}
solution here. 
\\end{solution}
\\fi
                            """
                        # elif path.endswith("_answers.tex"):
                        #     s = "\n".join(["\\item Option goes here." for _ in range(4) ])
                    else:
                        s = "Put your solution here."
                    f.write(s)
            # chk(sol)
            return path

        # sol =  chk("static/"+self.template_base +"_sol.tex")
        body = chk("static/"+self.template_base +"_body.tex")

        self.body = body

        return super().render(exam)

    def generate(self):
        x = SimpleNamespace()
        def txt2items(cmd):
            items = None
            if "\\begin{itemize}" in cmd:
                pre = cmd.split("\\begin{itemize}")[0]

                cmd = "x " + cmd.split("\\begin{itemize}")[1]
                cmd = cmd.split("\\end{itemize}")[0]
                items = cmd.split("\\item")[1:]
            else:
                pre = cmd

            return pre, items

        try:
            x.body = self.body
            with open(os.getcwd() + "/" + self.body, 'r') as f:
                ss = f.read()
            if "\\begin{solution}" in ss:
                x.text_solution = ss[ss.find("\\begin{solution}")+len("\\begin{solution}"):ss.find("\\end{solution}")]
                x.text_solution, x.solution_options = txt2items(x.text_solution)
            else:
                x.text_solution = None
                sfile = "static/"+self.template_base +"_body.tex"
                if os.path.isfile(sfile):
                    x.static_solution = sfile
                else:
                    raise Exception("No \\begin{solution} in template and no solution file: "+ sfile)


            cmd = ss.split("\\begin{comment}")[1]
            cmd = "x " + cmd.split("\\begin{itemize}")[1]
            cmd = cmd.split("\\end{itemize}")[0]
            cmd = cmd.split("\\item")[1:]
            if len(cmd) != 4:
                print(ss)
                print("="*100)
                raise Exception("Badly formed answers in: "+ self.body)
            x.answers = [a.strip() for a in cmd]


        except Exception as e:
            print("Bad generation", self.body)
        return x



import os, glob, shutil
import sys


def setup_irlc_files(exam):
    # TODO: Parameterize this to make it nicer.
    # Copy files from the irlc folder to the exam irlc folder.
    dirs = exam.get_dirs()
    # sbase = dirs.base + "/../../../02465public/src"
    # sbase = dirs.publ
    sbase = os.path.normpath(dirs.public_git_dir + "/src")
    assert os.path.isdir(sbase)

    dbase =  f"{dirs.base}/src"
    # public_root = os.path.normpath(dirs.base + "/../../../02465public")
    # print(public_root)
    for g in glob.glob(f"{dirs.public_git_dir}/src/irlc/**/*.py", recursive=True):
        srel = os.path.relpath(g, sbase)
        dabs = os.path.normpath(dbase + "/" + srel)
        if not os.path.isdir(os.path.dirname(dabs)):
            os.makedirs(os.path.dirname(dabs))
        # if "lectures" in g:
        #     print(os.path.normpath(g))
        shutil.copy(g, dabs)

def set_irlc_module_destination(irlc):
    # TODO: Parameterize this to make it nicer.

    import inspect
    frame = inspect.stack()[1]
    print("stack", inspect.stack())
    print("frame", frame)
    print("frame[0]", frame[0])
    try:
        module = inspect.getmodule(frame[0])
        _file = module.__file__
        __file__ = _file
        from pathlib import Path
        if Path(__file__).parent not in Path(irlc.__path__[0]).parents:
            sys.path = [p for p in sys.path if os.path.normpath(p) != os.path.dirname(irlc.__path__[0]) ]
            sys.path.append(os.path.dirname(__file__) + "/src")
        import importlib
        print("__file__", __file__)
        importlib.reload(irlc)
        assert Path(__file__).parent in Path(irlc.__path__[0]).parents
    except Exception as e:
        print("Got an exception", e)





from unitgrade import hide, UTestCase
from exam_generator.exam import Question, jinjafy
import numpy as np
import os
import inspect
from unitgrade import NotebookTestCase
from exam_generator.exam import Exam
import subprocess

JUPYTER_HEAD = """
---
jupytext:
  formats: ipynb,.pct.py:percent,.lgt.py:light,.spx.py:sphinx,md,Rmd,.pandoc.md:pandoc,.myst.md:myst
  text_representation:
    extension: '.md'
    format_name: myst
    format_version: '0.7'
    jupytext_version: 1.4.0+dev
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---
""".strip()

def get_jupyter_head(title='untitled document'):
    JUPYTER_HEAD = f"""
---
jupytext:
  formats: ipynb,.pct.py:percent,.lgt.py:light,.spx.py:sphinx,md,Rmd,.pandoc.md:pandoc,.myst.md:myst
  text_representation:
    extension: '.md'
    format_name: myst
    format_version: '0.7'
    jupytext_version: 1.4.0+dev
kernelspec:
  display_name: Python 3
  language: python
  name: python3
title: '{title}'
---
    """.strip()
    return JUPYTER_HEAD
    pass

class Mat1Exam(Exam):
    title = "Untiled problem set"

    def _get_base_location(self):
        base = os.path.dirname(inspect.getfile(self.__class__))
        return base

    def get_dirs(self):
        d = super().get_dirs()
        d.code = f"{d.code}_{self.seed}"
        if not os.path.isdir(d.code):
            os.makedirs(d.code)
        d.package_base = d.base +"/src"
        return d

    def render(self):
        def bracket(s):
            return "{"+ s+"}"
        # Take each question. Render it to an output directory. Then
        dirs = self.get_dirs()
        # qpath = []

        # questions = self.questions[0]['questions']
        # Resolve this to specific questions.

        qq = {}
        qq['questions'] = {}
        np.random.seed(self.seed)

        for k, q in self.questions.items():
            qq = self.questions[k]['question']
            self.questions[k]['variant'] = qq.questions[np.random.randint(len(qq.questions))]

        mb_base = """
{{JUPYTER_HEAD}}

{% for q in questions %}
{{q.md}}
{% endfor %}
        """.strip()

        x = {'mds': []}
        x['title'] = self.title
        x['JUPYTER_HEAD'] = get_jupyter_head(title=self.title)

        a =2
        for q in self.questions.values():
            # print("init", q)
            # if isinstance(q['question'], QuestionController):
            qv = q['variant'](seed=self.seed, question_args=q['question'].question_args)
            # else:
            #     qv = q['variant']

            # print("Didd init")
            x['mds'].append(qv.render(self))
            q['md'] = qv.render(self)

        x['questions'] = list(self.questions.values())
        md = jinjafy(dirs, mb_base, x)

        md_out = f"{dirs.code}/{self.name}.md"
        with open(md_out, "w") as f:
            f.write(md.strip())

        jupytext_file(md_out)

        def get_source(qq):
            source = []
            for k, v in qq.__dict__.items():
                if k not in ['sheet', 'generate', '__module__', '__annotations__', '__init__', 'build_md', 'name', 'notebook']:
                    print(k)
                    print(k, qq.__dict__[k])
                    if isinstance(qq.__dict__[k], bool) or v is None or isinstance(v, dict):
                        continue
                    s = inspect.getsource(qq.__dict__[k])
                    source.append(s)
            return source

        js = """
class Question{{n}}(NotebookTestCase):
    notebook = "{{notebook}}"
    
    {% for f in functions %}
{{f}}
    {% endfor %}    
"""
        qnames = []
        qs = []
        notebook = md_out[:-2] + "ipynb"
        for k, q in self.questions.items():
            functions = get_source(q['variant'])

            x = dict(notebook=os.path.basename(notebook), functions=functions, n=k + 1)
            qs.append( jinjafy(dirs, js, x))
            qnames.append(f"Question{k+1}")


        fulldoc = """
from unitgrade import hide, UTestCase
from unitgrade import UTestCase, Report
# from exam_generator.exam import Question, jinjafy
from types import SimpleNamespace
# import gym
import numpy as np
import sympy as sym
import os
import inspect
from unitgrade import NotebookTestCase

{% for q in qs %}
{{q}}
{% endfor %}

class {{report}}(Report):
    title = "{{report_title}}"
    abbreviate_questions = True
    questions = [
                {% for qn in qnames %}
                ({{qn}}, 10),
                {%- endfor %}
                ]    

    import {{package_name}}
    pack_imports = [{{package_name}}]


if __name__ == "__main__":
    from unitgrade import evaluate_report_student
    evaluate_report_student({{report}}())
"""
        name = self.name
        nname = self.base_name
        report = f"{nname[0].upper()}{nname[1:]}_{self.seed}"
        x = dict(qs=qs, report=report, qnames=qnames, report_title=f"Report for {nname}", package_name=self.package_name)
        s = jinjafy(dirs, fulldoc, x)
        print(s)
        out_complete = f"{dirs.code}/{name}_tests_complete.py"
        with open(out_complete, 'w') as f:
            f.write(s)

        deploy = """
from unitgrade_private import setup_grade_file_report
from snipper import snip_dir
from snipper.load_citations import get_aux, get_bibtex
{{ imporln }}

if __name__ == "__main__":
    setup_grade_file_report({{report}}, remove_hidden=False, bzip=False)
    setup_grade_file_report({{report}}, remove_hidden=True, bzip=False)    
        """
        imporln = f'from {os.path.relpath(out_complete, dirs.base + "/src").replace("/", ".")[:-3]} import {report}'
        x = dict(imporln=imporln, report=report)
        s = jinjafy(dirs, deploy, x)
        deploy_py = f"{dirs.code}/deploy.py"

        with open(deploy_py, 'w') as f:
            f.write(s)
        # Now deploy it.

        c = os.path.relpath( f"{dirs.code}/deploy.py", dirs.base + "/src").replace("/", ".")[:-3]
        src = dirs.base + "/src"
        cmd = f"cd {src} && python -m {c}"
        subprocess.run([cmd], shell=True)
        # Now this has been deployed. Deploy to a student repo.

def jupytext_file(md_file):
    d = os.path.dirname(md_file)
    md_file = os.path.abspath(md_file)
    assert os.path.isfile(md_file)

    cmd = f"cd {os.path.dirname(md_file)} && jupytext --to notebook --set-formats md:myst {os.path.basename(md_file)}"
    o = subprocess.run([cmd], shell=True, capture_output=True)
    print(str(o.stdout.decode("utf8")))

class QuestionController:
    questions = None
    def __init__(self, seed, question_args=None):
        self.seed = seed
        self.question_args = question_args

class Mat1Question(NotebookTestCase, Question):
    notebook = None
    sheet = None

    def __init__(self, *args, seed=None, question_args=None, **kwargs):
        # set up parts of the environment.
        if seed is None:
            seed = 0

        if question_args is None:
            question_args = {}
        Question.__init__(self, seed=seed, question_args=question_args)
        try:
            UTestCase.__init__(self, *args, **kwargs)
        except Exception as e:
            print("Bad init")
            raise e
    def render(self, exam):
        # Use the exam
        self.dirs = exam.get_dirs()
        self.exam = exam
        if not os.path.isdir(self.dirs.generated):
            os.makedirs(self.dirs.generated)
        if not os.path.isdir(self.dirs.figures):
            os.makedirs(self.dirs.figures)

        # Now it is appropriate for the great jinja.
        np.random.seed(self.seed)
        x = self.generate()
        if not isinstance(x, dict):
            x = x.__dict__
        x = {**x, **self._extra_generation(exam)}

        x['overall_exam_score'] = self.overall_exam_score
        x['question_title'] = self.question_title
        x['clearpage'] = self.clearpage
        x['include_solution_box'] = self.include_solution_box

        dn = os.path.basename(os.path.dirname(inspect.getfile(self.__class__)))
        texfile = os.path.dirname(inspect.getfile(self.__class__)) + "/" + os.path.basename(inspect.getfile(self.__class__))[:-3] + ".tex"

        # What should it do?
        # Render the question and return it as a str.
        # from
        s = jinjafy(self.dirs, self.sheet, x)
        return s

    def name(self):
        return f"{self.__class__.__name__}_{self.seed}"

    def render_dummy(self, r):

        class DummyExam(Mat1Exam):
            r = 23
            pass
        ex = DummyExam()
        s = self.render(ex)
        s = JUPYTER_HEAD + "\n\n"+s
        with open("./homework.md", 'w') as f:
            f.write(s.strip())
        # from exam_generator.mat1.mat1exam import jupytext_file
        jupytext_file("./homework.md")



from unitgrade import hide, UTestCase
from exam_generator.exam import Question, jinjafy
from types import SimpleNamespace
import gym
import numpy as np
import sympy as sym
import os
import inspect
from unitgrade import NotebookTestCase
from exam_generator.mat1.mat1exam import QuestionController, Mat1Question

# class Mat1Question(NotebookTestCase, Question):
#     notebook = None
#     sheet = None
#
#     def __init__(self, *args, seed=1, **kwargs):
#         # set up parts of the environment.
#         Question.__init__(self, seed=seed)
#         UTestCase.__init__(self, *args, **kwargs)
#
#     def render(self, exam):
#         # Use the exam
#         self.dirs = exam.get_dirs()
#
#         if not os.path.isdir(self.dirs.generated):
#             os.makedirs(self.dirs.generated)
#         if not os.path.isdir(self.dirs.figures):
#             os.makedirs(self.dirs.figures)
#
#         # Now it is appropriate for the great jinja.
#         np.random.seed(self.seed)
#         x = self.generate()
#         if not isinstance(x, dict):
#             x = x.__dict__
#         x = {**x, **self._extra_generation(exam)}
#
#         x['overall_exam_score'] = self.overall_exam_score
#         x['question_title'] = self.question_title
#         x['clearpage'] = self.clearpage
#         x['include_solution_box'] = self.include_solution_box
#
#         dn = os.path.basename(os.path.dirname(inspect.getfile(self.__class__)))
#         texfile = os.path.dirname(inspect.getfile(self.__class__)) + "/" + os.path.basename(
#             inspect.getfile(self.__class__))[:-3] + ".tex"
#
#         # What should it do?
#         # Render the question and return it as a str.
#         # from
#         s = jinjafy(self.dirs, self.sheet, x)
#         return s
#
#     def name(self):
#         return f"{self.__class__.__name__}_{self.seed}"
#
#     def render_dummy(self, r):
#         # Simplest?
#         # Make an exam and deploy to dummy file?
#         from exam_generator.mat1.mat1exam import Mat1Exam
#         class DummyExam(Mat1Exam):
#             r = 23
#             pass
#         ex = DummyExam()
#         s = self.render(ex)
#         with open("./homework.md", 'w') as f:
#             f.write(s.strip())
#         from exam_generator.mat1.mat1exam import jupytext_file
#         jupytext_file("./homework.md")

class VerificationQuestion(Mat1Question):
    # throw_error_on_no_tex_file = False
    notebook = 'homework.ipynb'
    sheet = """
# Hand-in code

The handin code will be written on the blackboard by the TA near the end of the test. Simply write the handin code below
(as a string) to confirm that you handed in during the test. When done, take a selfie of yourself and the handin code as a reciept, in case the 
handin code was mis-typed. 

```{code-cell} ipython3
handin_code = 'c000000'  # Write the handin code here. It has the format 'c123456789'.
handin_code = '{{handin_code}}' #!b;silent #!b
handin_code_checksum = sum([int(i) for i in handin_code[1:]])
if handin_code_checksum != {{handin_code_checksum}}:
    print("You specified the wrong handin code. Please ask the TA")
else:
    import subprocess, sys
    if "-nograde" not in sys.argv:
        subprocess.run(["cd ../../../ && python -m {{grade_script}} --noprogress --brief"], shell=True)        
```
"""
    def test_handin_code_checksum(self):
        self.assertEqualC(self.nb.handin_code_checksum)

    @hide
    def test_handin_code_checksum_hidden(self):
        self.assertEqualC(self.nb.handin_code_checksum)

    @hide
    def test_handin_code_hidden(self):
        self.assertEqualC(self.nb.handin_code)

    def generate(self):
        x = SimpleNamespace()
        x.a = 1+np.random.randint(100)
        x.b = 1+np.random.randint(100)
        print(self.question_args)
        print(self.dirs)
        x.handin_code = self.question_args['code']
        x.handin_code_checksum = sum([int(i) for i in x.handin_code[1:]])
        x.grade_script = f"{os.path.relpath(self.dirs.code, self.dirs.package_base).replace('/', '.')}.{self.exam.name}_grade"
        return x

class VerificationQuestionController(QuestionController):
    questions = [VerificationQuestion]



if __name__ == "__main__":
    # q = DP01(seed=3)
    # q.render_dummy(r = 3)
    # import unittest
    # unittest.main()
    pass


import setuptools
# py -m build && twine upload dist/*
# Linux> python -m build && python -m twine upload dist/*
# Local install: sudo pip install -e ./

# Manifest candidate:  include src/slider/DTU_Beamer_files/DTU_Beamer_files.zip
setuptools.setup(
    name="exam_generator",
    # packages=find_packages(),
    version="0.0.5",
    description="The DTU Exam generator for programming, conceptual and MC exams.",
    author="Tue Herlau",
    license="See licensing file",
    url='https://lab.compute.dtu.dk/tuhe/dtuexam', # require login; clean repo!
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=["jinja2", "matplotlib", "numpy"],
    include_package_data=True,
    package_data={'': ['templates/*'],},  # Check Manifest.in; required??
)



from unitgrade import hide, UTestCase
from exam_generator.exam import Question, jinjafy
from types import SimpleNamespace
import gym
import numpy as np
import sympy as sym
import os
import inspect
from unitgrade import NotebookTestCase
from exam_generator.mat1.mat1exam import QuestionController, Mat1Question


class DP01(Mat1Question):
    notebook = 'homework.ipynb'
    sheet = """
## A problem about adding numbers

### a) This problem will surely test your awesome number-adding skills.

In this problem, you should compute $x_1 = a + b$ where $a = {{a}}$ and $b = {{b}}$. 
```{code-cell} ipython3
x1 = None  # Write your result here.
x1 = {{a}} + {{b}} #!b;warn #!b
```

### b) Multiplication

In this problem, you should compute $x_2 = a b$ 
```{code-cell} ipython3
x2 = None  # Write your result here.
x2 = {{a * b}} #!b;warn #!b
```
"""
    def test_addition(self):
        self.assertIsNotNone(self.nb.x1)

    def test_multiplication(self):
        self.assertIsNotNone(self.nb.x2)

    @hide
    def test_addition_hidden(self):
        self.assertEqualC(self.nb.x1)

    @hide
    def test_multiplication_hidden(self):
        self.assertEqualC(self.nb.x2)

    def generate(self):
        x = SimpleNamespace()
        x.a = 1+np.random.randint(100)
        x.b = 1+np.random.randint(100)
        return x

class Arithmetics02(Mat1Question):
    # throw_error_on_no_tex_file = False
    notebook = 'homework.ipynb'
    sheet = """
## A problem about subtracting numbers.

This problem set will test your skills in subtracting numbers

### a) This problem will surely test your awesome number-adding skills.

In this problem, you should compute $x_1 = a + b - c$ where $a = {{a}}$ and $b = {{b}}$ and $c = {{c}}$ 
```{code-cell} ipython3
x1 = None  # Write your result here.
x1 = {{a + b - c}} #!b;warn #!b
print(x1)
```

### b) Multiplication and parenthesis

In this problem, you should compute $x_2 = a (b-c)$ 
```{code-cell} ipython3
x2 = None  # Write your result here.
x2 = {{a * (b-c) }} #!b;warn #!b
print(x2)
```
"""
    def test_addition(self):
        self.assertIsNotNone(self.nb.x1)
        self.assertEqual(self.nb.x_11, 4)

    def test_multiplication(self):
        self.assertIsNotNone(self.nb.x2)

    @hide
    def test_addition_hidden(self):
        self.assertEqualC(self.nb.x1)

    @hide
    def test_multiplication_hidden(self):
        self.assertEqualC(self.nb.x2)

    def generate(self):
        x = SimpleNamespace()
        x.a = 1+np.random.randint(20)
        x.b = 1+np.random.randint(10)
        x.c = 1 + np.random.randint(20)
        return x

class W1Q1Controller(QuestionController):
    questions = [DP01, Arithmetics02]

if __name__ == "__main__":
    q = DP01(seed=3)
    q.render_dummy(r = 3)
    import unittest
    unittest.main()



import shutil
import matplotlib.pyplot as plt
import jinja2
from types import SimpleNamespace
import numpy as np
import inspect
import os

class Exam:
    # Functional or OOP? Any difference?
    semester = 'spring'
    questions = []
    output_file = ''
    debug = False
    show_solutions = False
    date = 'March 4th, 2023'
    # centerlabel = 'Midterm test A'
    course_title = "Introduction to reinforcement learning and control"
    header_left = 'Midterm test A'
    header_center = '02465'
    duration = '4 hours'
    midterm = False

    permute_answers = True
    seed = None

    # Used to set package information for paths.
    package_name = 'irlc'
    subpackage_name = 'assignment1'

    @property
    def name(self):
        dirs = self.get_dirs()
        # name = os.path.basename(os.path.dirname(dirs.latex))
        return os.path.basename(dirs.code)

    @property
    def base_name(self):
        return os.path.basename(self._get_base_location())
        # return self.__module__

    def _get_base_location(self):

        print(self.__class__)
        print("Class.get_file(self.__class__", inspect.getfile(self.__class__) )

        gf =  inspect.getfile(self.__class__)
        abs_path = os.path.abspath(gf)
        print("abs_path", abs_path)
        inspect.getfile(self.__class__)
        ff = inspect.getfile(inspect.currentframe())
        print("using crurrent frame", ff)
        print("Using old method", gf)

        if "Exam" in gf:
            gf = gf
        elif "Exam" in abs_path:
            gf = abs_path
        else:
            gf = ff
        print("Returning base dirname from", gf)
        return os.path.dirname(gf)


    @property
    def out_path(self): # Unclear when/where this is used.
        return f"{self.package_name}/{self.subpackage_name}"

    def get_dirs(self):
        dirs = SimpleNamespace()
        base = self._get_base_location()
        print("The base name is", base)
        name = self.base_name
        # base = os.path.dirname(inspect.getfile(self.__class__))
        latex = base + "/latex"
        figures = base + "/latex/figures"

        dirs.base = base
        dirs.latex = latex
        dirs.figures = figures
        dirs.generated = base + "/latex/generated"
        dirs.code = f"{base}/src/{self.package_name}/{self.subpackage_name}/{name}"
        dirs.templates_base = os.path.dirname(inspect.getfile(Exam)) +"/templates"
        dirs.deployment = base + "/deployments"
        import questions as qs
        from exam_generator import questions as qs2

        dirs.templates_bases = [dirs.templates_base] +qs.__path__  + qs2.__path__ #  [os.path.dirname(p) for p in qs.__path__]
        dirs.templates_bases += [dirs.base]


        # This dir should be implemented in a special exam class that inherits from Exam. I do it here because it is simpler but a bit dirty.
        pdir = os.path.normpath(base + "/../../../02465public")
        if not os.path.isdir(pdir):
            pdir =  os.path.normpath(base + "/../../02465public")
            assert os.path.isdir(pdir)

        dirs.public_git_dir = pdir
        dirs.public_git_dir_relative_to_latex = os.path.relpath(dirs.public_git_dir, latex)
        return dirs

    def render(self):
        # If we are doing a deployment, remove the destionation file.

        def bracket(s):
            return "{"+ s+"}"
        # Take each question. Render it to an output directory. Then
        dirs = self.get_dirs()
        ext = '_solutions' if self.show_solutions else ''
        ext += '_noperm' if not self.permute_answers else ''
        file_out = f"{dirs.latex}/{self.name}{ext}.tex"
        deploy_to_pdf = dirs.deployment + "/" + os.path.basename(file_out)[:-4] + ".pdf"
        if os.path.isfile(deploy_to_pdf):
            os.remove(deploy_to_pdf)


        qpath = []

        # This construction is AFAIK not current anymore. Answers seem permuted just fine. Ignoring.
        if self.seed is None:
            pass
            # print("You must set an exam seed to permute the answers")
        np.random.seed(self.seed)

        for sec in self.questions:
            qpath.append(f"\section*{bracket(self.questions[sec]['title'])}")

            for q in self.questions[sec]['questions']:
                print(q)
                # print(q.program_file)
                file = q.render(self)
                path = os.path.relpath(file, dirs.latex)
                path = path.replace("\\", "/")
                qpath.append(f"\input{bracket(path)}")

        pydest = f"{dirs.code}/multiple_choice_answers.py"
        x = {'questions': qpath, 'show_frontpage': True, 'course_title': self.course_title, 'exam_name': self.name, 'date': self.date,
             'midterm_base': os.path.relpath(dirs.code, dirs.base+"/src"),
             'duration': self.duration, 'midterm': self.midterm,
             'header_left': self.header_left,
             'header_center': self.header_center,
             # 'centerlabel': self.centerlabel,
             'show_solutions': self.show_solutions,
             'N_MC': len(self.questions[0]['questions']),
             'mc_answer_path':         os.path.relpath(pydest, dirs.base+"/src"),
             'programming_handin_files': [os.path.relpath(dirs.code, dirs.base+"/src") + "/" + q.program_file for q in self.questions[2]['questions']] ,
             'gradepy': os.path.relpath(dirs.code, dirs.base+"/src") + f'/{self.name}_tests_grade.py',
             'dirs': dirs,
             }

        jinja_write(dirs, template_base="exam_base.tex", x=x, file_out=file_out)
        from slider.latexutils import latexmk
        try:
            pdf_out = file_out[:-4] + ".pdf"
            if os.path.isfile(pdf_out):
                os.remove(pdf_out)
            # Remove the PDF file.
            # print("Compiling latex...")
            latexmk(file_out)
            # print("Compiled.")
            if not self.show_solutions and self.permute_answers:
                # Deployment of exam.
                # pdf_dest = dirs.deployment + "/"+os.path.basename(pdf_out)
                shutil.copyfile(file_out[:-4]+".pdf", deploy_to_pdf)


        except Exception as e:
            with open(file_out[:-4]+".log") as f:
                print(f.read())
            print("Exception caught and handled...")
            raise Exception("The PDF compilation fucked up. I was compiling: ", file_out)


        """ Make answer files for MC questions. """

        # pyans = os.path.dirname(__file__ ) + "/templates/multiple_choice_answers.py"
        # pydest =  f"{dirs.base}/src/{os.path.basename(dirs.base )}/multiple_choice_answers.py"
        x = {'NQ': len(self.questions[0]['questions'])}

        jinja_write(dirs, template_base='multiple_choice_answers.pyt', x=x, file_out=pydest)


def jinja_write(dirs, x, template_base=None, template_str=None, file_out=None):
    templateLoader = jinja2.FileSystemLoader(searchpath=dirs.templates_bases)
    templateLoader = jinja2.FileSystemLoader(searchpath=dirs.templates_bases)

    env = jinja2.Environment(lstrip_blocks=True, trim_blocks=True, loader=templateLoader)
    from exam_generator.jinjaenv import latex_env
    env = latex_env(env)

    if template_base is not None:
        # print(template_base)
        rp = [os.path.relpath(template_base, r) for r in dirs.templates_bases]
        rp = [r for r in rp if not r.startswith("..")]
        if len(rp) == 0:
            rp = [template_base]
        # if len(rp) == 0:
        #     print("Bad! ")
        #     print(template_base, dirs.templates_bases)
        #     assert False
        try:
            template = env.get_template(rp[0])
        except Exception as e:
            print(e)
            raise e
    else:
        template = env.from_string(template_str)
    jinja_out = template.render(x)
    if file_out is not None:
        with open(file_out, 'w', encoding='utf-8') as f:
            f.write(jinja_out)
            # print("Writing to: " + file_out)
    return jinja_out



def jinjafy(dirs, jinja_str, x):
    templateLoader = jinja2.FileSystemLoader(searchpath=dirs.templates_bases)
    env = jinja2.Environment(lstrip_blocks=True, trim_blocks=True, loader=templateLoader)
    from exam_generator.jinjaenv import latex_env
    env = latex_env(env)
    template = env.from_string(jinja_str)
    # template = env.get_template(template_base)
    jinja_out = template.render(x)
    return jinja_out

class Question:
    # template = ''

    dirs = None # set by the exam.
    seed = 3
    _number = 0
    overall_exam_score = 1
    question_title = ''
    include_solution_box = False
    throw_error_on_no_tex_file = True

    template = None

    def __init__(self, seed, *args, title='', clearpage=False, question_args=None):
        self.seed = seed

        self.question_title = title
        self.clearpage = clearpage
        self.question_args = question_args

    def setup(self, args, **kwargs):
        x = SimpleNamespace()
        x.foo = 'bar'
        return x

    def generate(self):
        # Should actually generate the output of this question. This is the method you are supposed to implement.
        raise NotImplementedError("Implement this.")

    def _extra_generation(self, exam):
        return {}

    def render(self, exam):
        # Use the exam
        self.dirs = exam.get_dirs()

        if not os.path.isdir(self.dirs.generated):
            os.makedirs(self.dirs.generated)
        if not os.path.isdir(self.dirs.figures):
            os.makedirs(self.dirs.figures)

        # Now it is appropriate for the great jnja.
        np.random.seed(self.seed)
        self.rng = np.random.default_rng(self.seed)
        # print(self)
        x = self.generate()
        if not isinstance(x, dict):
            x = x.__dict__
        x = {**x, **self._extra_generation(exam)}

        x['overall_exam_score'] = self.overall_exam_score
        x['question_title'] = self.question_title
        x['clearpage'] = self.clearpage
        x['include_solution_box'] = self.include_solution_box

        dn = os.path.basename(os.path.dirname(inspect.getfile(self.__class__)))
        if self.template is None:
            texfile = os.path.dirname(inspect.getfile(self.__class__)) + "/" + os.path.basename(inspect.getfile(self.__class__))[:-3] + ".tex"
            if not os.path.exists(texfile):
                print("> Texfile for question not found:", texfile)
                import glob

                fls = glob.glob(os.path.dirname(texfile) + "/*.tex")
                if len(fls) == 1:
                    print("> Attempting to move alternative texfile in same directory", fls[0])
                    import shutil
                    shutil.move(fls[0], texfile)
                else:
                    if self.throw_error_on_no_tex_file:
                        raise Exception("Texfile not found " + texfile)
                print("Textfile not found", texfile)
            # rtex = os.path.relpath(texfile, self.dirs.templates_bases[1])
            # rtex = rtex.replace("\\", "/")
            texfile = texfile.replace("\\", "/")
        else:
            texfile = None

        file_out = self.dirs.generated + "/" + dn + f"_{Question._number}.tex"
        if os.path.isfile(texfile):
            jinja_write(self.dirs,x=x, template_base=texfile, template_str=self.template, file_out=file_out)
            Question._number += 1
            return file_out
        return None
    def savepdf(self, filename):
        if filename.endswith(".pdf"):
            raise Exception("Please remove the .pdf extension when calling savepdf. Your filename:", filename)

        if not filename.endswith(".pdf"):
            filename += ".pdf"
        filein = self.dirs.latex +"/figures/"+filename
        plt.savefig(self.dirs.latex +"/figures/"+filename, dpi = 250) # default 100.
        from slider.convert import pdfcrop
        pdfcrop(filein, filein)


class ProgrammingQuestion(Question):
    program_file = "PLEAST_SET_PROGRAM_FILE.py"
    include_solution_box = True

    def render(self, exam):
        return super().render(exam)


    def _extra_generation(self, exam):
        dirs = exam.get_dirs()
        with open(dirs.code + "/"+self.program_file, 'r') as f:
            s = f.read()
        lines = s.splitlines()
        x = {}
        for l in lines:
            if l.startswith("def ") and l[5] == "_" and l[4] in 'abcdefg':
                fname = l
                if "->" in fname:
                    fname = fname.split("->")[0].strip() +":"
                x['problem_' + l[4]] = fname[:-1]


        return {'program_file':self.program_file,
                'program_path': os.path.relpath(dirs.code, dirs.base+"/src") + "/"+self.program_file,
                'solution_head': 'The solution can be found in this directory as a \\pyi{.py} file.',
                **x}

class MCQuestion(Question):
    include_solution_box = True

    def __init__(self, *args, answer_permutation=None, **kwargs):
        self.answer_permutation = answer_permutation
        super().__init__(*args, **kwargs)

    def _extra_generation(self, exam):
        # if hasattr(self, 'template_base') and self.template_base == 'bellman': # self.seed == 204:
        #     print("here")
        x = super()._extra_generation(exam)
        if exam.permute_answers:
            # self.set_answer_permutation()
            if self.answer_permutation is not None:
                pass
            else:
                self.answer_permutation = self.rng.permutation(4)
        else:
            self.answer_permutation = (0,1,2,3)

        assert len(self.answer_permutation) == 4
        x['correct_answer'] = 'ABCD'[list(self.answer_permutation).index(0)]

        x['answer_permutation'] = self.answer_permutation
        # print("Using permutation", self.answer_permutation)
        # raise Exception(f"Rng state is {self.rng}")
        return x

    # def set_answer_permutation(self, permutation=None):
    #     # np.random.seed(self.seed)
    #     if permutation is None:
    #         assert False
    #         self.answer_permutation = self.rng.permutation(4)
    #         # print("Using permutation", self.answer_permutation)
    #     else:
    #         self.answer_permutation = permutation